@@ -2,7 +2,7 @@ package HTML::Selector::XPath;
 
 use strict;
 use 5.008_001;
-our $VERSION = '0.18';
+our $VERSION = '0.23';
 
 require Exporter;
 our @EXPORT_OK = qw(selector_to_xpath);
@@ -28,7 +28,7 @@ my $reg = {
         ( [~|*^\$!]? = ) \s*
         (?: ($ident) | "([^"]*)" | '([^']*)') \s* \] /x,
     badattr => qr/^\[/,
-    attrN   => qr/^:not\((.*?)\)/i, # this should be a parentheses matcher instead of a RE!
+    attrN   => qr/^:not\(/i, # we chop off the closing parenthesis below in the code
     pseudo  => qr/^:([()a-z0-9_+-]+)/i,
     # adjacency/direct descendance
     combinator => qr/^(\s*[>+~\s](?!,))/i,
@@ -97,12 +97,14 @@ sub nth_last_child {
     _generate_child('following', $a, $b);
 }
 
-sub to_xpath {
-    my $self = shift;
-    my $rule = $self->{expression} or return;
-    my %parms = @_;
+# A hacky recursive descent
+# Only descends for :not(...)
+sub consume {
+    my ($self, $rule, %parms) = @_;
     my $root = $parms{root} || '/';
 
+    return [$rule,''] if $rule =~ m!^/!; # If we start with a slash, we're already an XPath?!
+
     my @parts = ("$root/");
     my $last_rule = '';
     my @next_parts;
@@ -165,18 +167,24 @@ sub to_xpath {
 
         # Match negation
         if ($rule =~ s/$reg->{attrN}//) {
-            my $sub_rule = $1;
-            if ($sub_rule =~ s/$reg->{attr2}//) {
+            # Now we parse the rest, and after parsing the subexpression
+            # has stopped, we must find a matching closing parenthesis:
+            if ($rule =~ s/$reg->{attr2}//) {
                 push @parts, "[not(", convert_attribute_match( $1, $2, $^N ), ")]";
-            } elsif ($sub_rule =~ s/$reg->{attr1}//) {
+            } elsif ($rule =~ s/$reg->{attr1}//) {
                 push @parts, "[not(\@$1)]";
-            } elsif ($sub_rule =~ /$reg->{badattr}/) {
-                Carp::croak "Invalid negated attribute-value selector ':not($sub_rule)'";
+            } elsif ($rule =~ /$reg->{badattr}/) {
+                Carp::croak "Invalid negated attribute-value selector ':not($rule)'";
             } else {
-                my $xpath = selector_to_xpath($sub_rule);
-                $xpath =~ s!^//!!;
+                my( $new_parts, $leftover ) = $self->consume( $rule, %parms );
+                shift @$new_parts; # remove '//'
+                my $xpath = join '', @$new_parts;
+                
                 push @parts, "[not(self::$xpath)]";
+                $rule = $leftover;
             }
+            $rule =~ s!^\s*\)!!
+                or die "Unbalanced parentheses at '$rule'";
         }
 
         # Ignore pseudoclasses/pseudoelements
@@ -209,9 +217,14 @@ sub to_xpath {
             } elsif ($1 =~ /^nth-of-type\((\d+)\)$/) {
                 push @parts, "[$1]";
             } elsif ($1 =~ /^contains\($/) {
-                $rule =~ s/^\s*"([^"]*)"\s*\)//
-                    or die "Malformed string in :contains(): '$rule'";
-                push @parts, qq{[text()[contains(string(.),"$1")]]};
+                if( $rule =~ s/^\s*"([^"]*)"\s*\)// ) {
+                    push @parts, qq{[text()[contains(string(.),"$1")]]};
+                } elsif( $rule =~ s/^\s*'([^']*)'\s*\)// ) {
+                    push @parts, qq{[text()[contains(string(.),"$1")]]};
+                } else {
+                    return( \@parts, $rule );
+                    #die "Malformed string in :contains(): '$rule'";
+                };
             } elsif ( $1 eq 'root') {
                 # This will give surprising results if you do E > F:root
                 $parts[$root_index] = $root;
@@ -248,7 +261,19 @@ sub to_xpath {
             undef $wrote_tag;
         }
     }
-    return join '', @parts;
+    return \@parts, $rule
+}
+
+sub to_xpath {
+    my $self = shift;
+    my $rule = $self->{expression} or return;
+    my %parms = @_;
+    
+    my($result,$leftover) = $self->consume( $rule, %parms );
+    $leftover
+        and die "Invalid rule, couldn't parse '$leftover'";
+    return join '', @$result;
+
 }
 
 sub parse_pseudo { 
